@@ -7,11 +7,11 @@ import os
 # 3rd-party
 from lxml import etree
 
-_TSFMT = "%Y%m%d%H%M%S"
-
 import json
 from jnpr.junos.factory.to_json import TableJSONEncoder
 
+_TSFMT = "%Y%m%d%H%M%S"
+
 
 class Table(object):
     ITEM_XPATH = None
@@ -126,9 +126,11 @@ class Table(object):
         key_value, xpath = self._keyspec()
 
         if isinstance(key_value, str):
-            # Check if pipe is in the key_value, if so append xpath to each value
+            # Check if pipe is in the key_value, if so append xpath
+            # to each value
             if ' | ' in key_value:
-                return self._keys_simple(' | '.join([xpath + '/' + x for x in key_value.split(' | ')]))
+                return self._keys_simple(' | '.join([xpath + '/' + x for x in
+                                                     key_value.split(' | ')]))
             return self._keys_simple(xpath + '/' + key_value)
 
         if not isinstance(key_value, list):
